from lingua import Language

language_codes = {
    "AFRIKAANS": 'af',
    "ALBANIAN": 'sq',
    "ARABIC": 'ar',
    "ARMENIAN": 'hy',
    "AZERBAIJANI": 'az',
    "BASQUE": 'eu',
    "BELARUSIAN": 'be',
    "BENGALI": 'bn',
    "BOKMAL": 'nb',
    "BOSNIAN": 'bs',
    "BULGARIAN": 'bg',
    "CATALAN": 'ca',
    "CHINESE": 'zh',
    "CROATIAN": 'hr',
    "CZECH": 'cs',
    "DANISH": 'da',
    "DUTCH": 'nl',
    "ENGLISH": 'en',
    "ESPERANTO": 'eo',
    "ESTONIAN": 'et',
    "FINNISH": 'fi',
    "FRENCH": 'fr',
    "GANDA": 'lg',
    "GEORGIAN": 'ka',
    "GERMAN": 'de',
    "GREEK": 'el',
    "GUJARATI": 'gu',
    "HEBREW": 'he',
    "HINDI": 'hi',
    "HUNGARIAN": 'hu',
    "ICELANDIC": 'is',
    "INDONESIAN": 'id',
    "IRISH": 'ga',
    "ITALIAN": 'it',
    "JAPANESE": 'ja',
    "KAZAKH": 'kk',
    "KOREAN": 'ko',
    "LATIN": 'la',
    "LATVIAN": 'lv',
    "LITHUANIAN": 'lt',
    "MACEDONIAN": 'mk',
    "MALAY": 'ms',
    "MAORI": 'mi',
    "MARATHI": 'mr',
    "MONGOLIAN": 'mn',
    "NYNORSK": 'nn',
    "PERSIAN": 'fa',
    "POLISH": 'pl',
    "PORTUGUESE": 'pt',
    "PUNJABI": 'pa',
    "ROMANIAN": 'ro',
    "RUSSIAN": 'ru',
    "SERBIAN": 'sr',
    "SHONA": 'sn',
    "SLOVAK": 'sk',
    "SLOVENE": 'sl',
    "SOMALI": 'so',
    "SOTHO": 'st',
    "SPANISH": 'es',
    "SWAHILI": 'sw',
    "SWEDISH": 'sv',
    "TAGALOG": 'tl',
    "TAMIL": 'ta',
    "TELUGU": 'te',
    "THAI": 'th',
    "TSONGA": 'ts',
    "TSWANA": 'tn',
    "TURKISH": 'tr',
    "UKRAINIAN": 'uk',
    "URDU": 'ur',
    "VIETNAMESE": 'vi',
    "WELSH": 'cy',
    "XHOSA": 'xh',
    "YORUBA": 'yo',
    "ZULU": 'zu'
}